#define _USE_MATH_DEFINES
 
#include "wViscFluxes.h"
#include "wBLRegion.h"
#include <iostream>
#include <Eigen/Dense>
#include <cmath>
#include <algorithm>
#include <string>

using namespace dart;
using namespace Eigen;


ViscFluxes::ViscFluxes()
{
  std::cout << "Coucou de ViscFluxes\n" << std::endl;

  
} // end ViscFluxes


ViscFluxes::~ViscFluxes()
{
    std::cout << "~ViscFluxes()\n";
} // end ~ViscFluxes

Vector<double, 5> ViscFluxes::ComputeFluxes(size_t iPoint, BLRegion *bl)
{
  unsigned int nVar = bl->GetnVar();
  std::vector<double> u;
  for (size_t k=0; k<nVar; ++k){
    u.push_back(bl->U[iPoint*nVar + k]);
  } // End for
  return BLlaws(iPoint, bl, u);
}

Matrix<double, 5, 5> ViscFluxes::ComputeJacobian(size_t iPoint, BLRegion *bl, Vector<double, 5> &sysRes, double eps)
{
  unsigned int nVar = bl->GetnVar();
  Matrix<double, 5, 5> JacMatrix;
  std::vector<double> uUp;
  for (size_t k=0; k<nVar; ++k){
    uUp.push_back(bl->U[iPoint*nVar + k]);
  } // End for

  for (size_t iVar=0; iVar<nVar; ++iVar){
      double varSave = uUp[iVar];
      uUp[iVar] += eps;

      JacMatrix.col(iVar) = (BLlaws(iPoint, bl, uUp) - sysRes) / eps;

      uUp[iVar] = varSave;
  }
  return JacMatrix;
}

Vector<double, 5> ViscFluxes::BLlaws(size_t iPoint, BLRegion *bl, std::vector<double> u)
{
  unsigned int nVar = bl->GetnVar();

  Matrix<double, 5, 5> timeMatrix;
  Vector<double, 5> spaceVector;

  double dissipRatio;
  if (bl->name == "wake"){
    double dissipRatio = 0.9;
  } // End if
  else{
    double dissiRatio = 1;
  } // End else

  bl->Ret[iPoint] = std::max(u[3] * u[0], 1.0);
  bl->DeltaStar[iPoint] = u[0] * u[1];

  /* Boundary layer closure relations */

  if (bl->Regime[iPoint] == 0){
    LaminarClosures(iPoint, bl);
  } // End if
  else if (bl->Regime[iPoint] == 1){
    TurbulentClosures(iPoint, bl);
  } // End else if

  else{
    std::cout << "Wrong regime\n" << std::endl;
  } // End else

  /* Gradients computation */
  double dx = (bl->GetLocMarkers(iPoint) - bl->GetLocMarkers(iPoint-1));
  double dT_dx = (u[0] - bl->U[(iPoint-1)*nVar + 0])/dx;
  double dH_dx = (u[1] - bl->U[(iPoint-1)*nVar + 1])/dx;
  double due_dx = (u[3] - bl->U[(iPoint-1)*nVar + 3])/dx;
  double dCt_dx = (u[4] - bl->U[(iPoint-1)*nVar + 4])/dx;
  double dHstar_dx = (bl->Hstar[iPoint] - bl->Hstar[iPoint-1])/dx;

  double dueExt_dx = (bl->GetVtExt(iPoint) - bl->GetVtExt(iPoint))/dx;
  double ddeltaStarExt_dx = (bl->GetDeltaStarExt(iPoint) - bl->GetDeltaStarExt(iPoint))/dx;

  double c=4/(M_PI*dx);
  double cExt=4/(M_PI*(bl->GetXxExt(iPoint) - bl->GetXxExt(iPoint-1)));

  /* Amplification routine */
  double dN_dx;
  double Ax;

  if (bl->xtrF!=NULL && bl->Regime[iPoint] == 0){
    double Ax = AmplificationRoutine(bl->Hk[iPoint], bl->Ret[iPoint], u[0]);
    double dN_dx = (bl->U[iPoint * nVar + 2] - bl->U[(iPoint-1)*nVar + 2])/(bl->GetLocMarkers(iPoint) - bl->GetLocMarkers(iPoint-1));

  } // End if
  else{
    double Ax = 0;
    double dN_dx = 0;
  } // End else

  double Mea = bl->GetMe(iPoint);

  double Reta = bl->Ret[iPoint];
  double Hstara = bl->Hstar[iPoint];
  double Hstar2a = bl->Hstar2[iPoint];
  double Hka = bl->Hk[iPoint];
  double Cfa = bl->Cf[iPoint];
  double Cda = bl->Cd[iPoint];
  double deltaa = bl->Delta[iPoint];
  double Usa = bl->Us[iPoint];

  /* Space part */

  spaceVector(0) = dT_dx + (2 + u[1] - Mea * Mea) * (u[0]/u[3]) * due_dx - Cfa;
  spaceVector(1) = u[1] * dHstar_dx + (2*Hstar2a + Hstara * (1-u[1])) * u[0]/u[3] * due_dx - 2*Cda + Hstara * Cfa/2;
  spaceVector(2) = dN_dx - Ax;
  spaceVector(3) = due_dx - c*(u[1] * dT_dx + u[0]*dH_dx) - dueExt_dx + cExt * ddeltaStarExt_dx;

  if (bl->Regime[iPoint] == 1){
    spaceVector(4) = (2*deltaa/u[4]) * dCt_dx - 5.6*(bl->Cteq[iPoint] - u[4]*dissipRatio)
                      - 2*deltaa*(4/(3*u[1]*u[0]) * (Cfa/2 - ((Hka-1)/(6.7*Hka * dissipRatio)) * ((Hka-1)/(6.7*Hka * dissipRatio))) - 1/u[3] * due_dx);
  } // End if
  else{
    spaceVector(4) = 0;
  }

  /* Time part */

  timeMatrix(0,0) = u[1]/u[3];
  timeMatrix(0,1) = u[0]/u[3];
  timeMatrix(0,2) = 0;
  timeMatrix(0,3) = u[0]*u[1]/(u[3]*u[3]);
  timeMatrix(0,4) = 0;

  timeMatrix(1,0) = (1+u[1]*(1-Hstara))/u[3];
  timeMatrix(1,1) = (1-Hstara)*u[1]/u[3];
  timeMatrix(1,2) = 0;
  timeMatrix(1,3) = (2-Hstara*u[1])*u[0]/(u[3]*u[3]);
  timeMatrix(1,4) = 0;

  timeMatrix(2,0) = 0;
  timeMatrix(2,1) = 0;
  timeMatrix(2,2) = 1;
  timeMatrix(2,3) = 0;
  timeMatrix(2,4) = 0;

  timeMatrix(3,0) = -c*u[1];
  timeMatrix(3,1) = -c*u[0];
  timeMatrix(3,2) = 0;
  timeMatrix(3,3) = 1;
  timeMatrix(3,4) = 0;
  
  timeMatrix(4,0) = 0;
  timeMatrix(4,1) = 0;
  timeMatrix(4,2) = 0;

  if (bl->Regime[iPoint]==1){
    timeMatrix(4,3) = 2*deltaa/(Usa*u[3]*u[3]);
    timeMatrix(4,4) = 2*deltaa/(u[3]*u[4]*Usa);
  }
  else{
    timeMatrix(4,3) = 0;
    timeMatrix(4,4) = 0;
  }

  Vector<double, 5> sysRes = timeMatrix.partialPivLu().solve(spaceVector);
  return sysRes;
}

double ViscFluxes::AmplificationRoutine(double Hk, double Ret, double theta)
{
  double Ax;

  double Dgr = 0.08;
  double Hk1 = 3.5;
  double Hk2 = 4;
  double Hmi = (1/(Hk-1));
  double logRet_crit = 2.492*pow(1/(Hk-1),0.43) + 0.7*(tanh(14*Hmi-9.24)+1.0); // Value at which the amplification starts to grow
  if (Ret <=0){
      Ret = 1;
  }    
  if (log10(Ret) < logRet_crit - Dgr){
      Ax = 0;
  }
  else{
      double Rnorm = (log10(Ret) - (logRet_crit-Dgr))/(2*Dgr);
      double Rfac;
      if(Rnorm <=0){
          Rfac = 0;
      }
      else if(Rnorm >= 1){
          Rfac = 1;
      }
      else{
          Rfac = 3*Rnorm*Rnorm - 2*Rnorm*Rnorm*Rnorm;
      }
      // Normal envelope amplification rate
      double m = -0.05+2.7*Hmi-5.5*Hmi*Hmi+3*Hmi*Hmi*Hmi+0.1*exp(-20*Hmi);
      double arg = 3.87*Hmi-2.52;
      double dndRet = 0.028*(Hk-1)-0.0345*exp(-(arg*arg));
      Ax = (m*dndRet/theta)*Rfac;
      // Set correction for separated profiles
      if(Hk > Hk1){
          double Hnorm = (Hk - Hk1)/(Hk2-Hk1);
          double Hfac;
          if(Hnorm >=1){
              Hfac = 1;
          }
          else{
              Hfac = 3*Hnorm*Hnorm - 2*Hnorm*Hnorm*Hnorm;
          }
          double Ax1 = Ax;
          double Gro = 0.3+0.35*exp(-0.15*(Hk-5));
          double Tnr = tanh(1.2*(log10(Ret)-Gro));
          double Ax2 = (0.086*Tnr - 0.25/pow(Hk-1,1.5))/theta;
          if(Ax2 < 0){
              Ax2 = 0;
          }
          Ax = Hfac*Ax2 + (1-Hfac)*Ax1;
          Ax = std::max(Ax, 0.0);
      }
  }
  return Ax;
}

void ViscFluxes::LaminarClosures(size_t iPoint, BLRegion *bl)
{ 
  unsigned int nVar = bl->GetnVar();
  double theta = bl->U[iPoint*nVar+0];
  double H = bl->U[iPoint*nVar+1];
  double Ret = bl->Ret[iPoint];
  double Me = bl->GetMe(iPoint);
  double Hk = (H - 0.29*Me*Me)/(1+0.113*Me*Me); // Kinematic shape factor
  if(bl->name == "airfoil"){
      Hk = std::max(Hk, 1.02);
  }
  else{
      Hk = std::max(Hk, 1.00005);
  }
  Hk = std::min(Hk,6.0);
  double Hstar2 = (0.064/(Hk-0.8)+0.251)*Me*Me; // Density shape parameter
  double delta = std::min((3.15+ H + (1.72/(Hk-1)))*theta, 12*theta);
  double Hks = Hk - 4.35;
  double Hstar;
  if(Hk <= 4.35){
      double dens = Hk + 1;
      Hstar = 0.0111*Hks*Hks/dens - 0.0278*Hks*Hks*Hks/dens + 1.528 -0.0002*(Hks*Hk)*(Hks*Hk);
  }
  else if(Hk > 4.35){
      Hstar = 1.528 + 0.015*Hks*Hks/Hk;
  }
  Hstar = (Hstar + 0.028*Me*Me)/(1+0.014*Me*Me); // Whitfield's minor additional compressibility correction

  /* Cf */

  double Cf;
  if(Hk < 5.5){
      double tmp = (5.5-Hk)*(5.5-Hk)*(5.5-Hk) / (Hk+1);
      Cf = (-0.07 + 0.0727*tmp)/Ret;
  }
  else if(Hk >= 5.5){
      double tmp = 1 - 1/(Hk-4.5);
      Cf = (-0.07 + 0.015*tmp*tmp) /Ret;
  }

  /* Cd */

  double Cd;
  if(Hk < 4){
      Cd = ( 0.00205*pow(4.0-Hk, 5.5) + 0.207 ) * (Hstar/(2*Ret));
  }
  else if(Hk >= 4){
      double HkCd = (Hk-4)*(Hk-4);
      double denCd = 1+0.02*HkCd;
      Cd = ( -0.0016*HkCd/denCd + 0.207) * (Hstar/(2*Ret));
  }
  if (bl->name == "wake"){
      Cd = 2*(1.10*(1.0-1.0/Hk)*(1.0-1.0/Hk)/Hk)* (Hstar/(2*Ret));
      Cf = 0;
  }
  bl->Cd[iPoint] = Cd;
  bl->Cf[iPoint] = Cf;
  bl->Delta[iPoint] = delta;
  bl->Hstar[iPoint] = Hstar;
  bl->Hstar2[iPoint] = Hstar2;
  bl->Hk[iPoint] = Hk;
  bl->Cteq[iPoint] = 0;
  bl->Us[iPoint] = 0;
}

void ViscFluxes::TurbulentClosures(size_t iPoint, BLRegion *bl)
{
  unsigned int nVar = bl->GetnVar();
  double theta = bl->U[iPoint*nVar+0];
  double H = bl->U[iPoint*nVar+1];
  double Ct = bl->U[iPoint*nVar+4];
  double Ret = bl->Ret[iPoint];
  double Me = bl->GetMe(iPoint);

  H = std::max(H, 1.0005);
  /* Eliminate absurd transients */
  Ct = std::max(std::min(Ct, 0.30), 0.0000001);
  double Hk = (H - 0.29*Me*Me)/(1+0.113*Me*Me);
  Hk = std::max(std::min(Hk, 10.0), 1.00005);
  double Hstar2 = ((0.064/(Hk-0.8))+0.251)*Me*Me;
  //gam = 1.4 - 1
  //Fc = np.sqrt(1+0.5*gam*Me**2)
  double Fc = sqrt(1+0.2*Me*Me);
  
  double H0;
  if(Ret <= 400){
    H0 = 4;
  }
  else if(Ret > 400){
    H0 = 3 + (400/Ret);
  }

  Ret = std::max(Ret, 200.0);
  double Hstar;
  if (Hk <= H0){
      Hstar = ((0.5-4/Ret)*((H0-Hk)/(H0-1))*((H0-Hk)/(H0-1)))*(1.5/(Hk+0.5))+1.5+4/Ret;
  }
  else if (Hk > H0){
      Hstar = (Hk-H0)*(Hk-H0)*(0.007*log(Ret)/((Hk-H0+4/log(Ret))*(Hk-H0+4/log(Ret))) + 0.015/Hk) + 1.5 + 4/Ret;
  }

  Hstar = (Hstar + 0.028*Me*Me)/(1+0.014*Me*Me); // Whitfield's minor additional compressibility correction
  //logRt = np.log(Ret/Fc)
  //logRt = max(logRt,3)
  //arg = max(-1.33*Hk, -20)
  double Us = (Hstar/2)*(1-4*(Hk-1)/(3*H)); // Equivalent normalized wall slip velocity
  double delta = std::min((3.15+ H + (1.72/(Hk-1)))*theta, 12*theta);
  double Ctcon = 0.5/(6.7*6.7*0.75);

  int n;
  double Cf;
  double Hkc;
  double Cdw;
  double Cdd;
  double Cdl;
  double Cteq;
  double Hmin;
  double Fl;
  double Dfac;

  if (bl->name == "wake"){
      Us = std::min(Us, 0.99995);
      n = 2; // two wake halves
      Cf = 0; // no friction inside the wake
      Hkc = Hk - 1;
      Cdw = 0;  // wall contribution
      Cdd = (0.995-Us)*Ct*Ct; // turbulent outer layer contribution
      Cdl = 0.15*(0.995-Us)*(0.995-Us)/Ret; // laminar stress contribution to outer layer
      Cteq = sqrt(4*Hstar*Ctcon*((Hk-1)*Hkc*Hkc)/((1-Us)*(Hk*Hk)*H)); // Here it is Cteq^0.5
  }
  else{
      if(Us > 0.95){
          Us = 0.98;
      } // End if 
      n = 1;
      // Cf = (1/Fc)*(0.3*np.exp(arg)*(logRt/2.3026)**(-1.74-0.31*Hk)+0.00011*(np.tanh(4-(Hk/0.875))-1))
      Cf = 1/Fc*(0.3*exp(-1.33*Hk)*pow(log10(Ret/Fc),-1.74-0.31*Hk) + 0.00011*(tanh(4-Hk/0.875) - 1));
      Hkc = std::max(Hk-1-18/Ret, 0.01);
      // Dissipation coefficient
      Hmin = 1 + 2.1/log(Ret);
      Fl = (Hk-1)/(Hmin-1);
      Dfac = 0.5+0.5*tanh(Fl);
      Cdw = 0.5*(Cf*Us) * Dfac;
      Cdd = (0.995-Us)*Ct*Ct;
      Cdl = 0.15*(0.995-Us)*(0.995-Us)/Ret;
      Cteq = sqrt(Hstar*Ctcon*((Hk-1)*Hkc*Hkc)/((1-Us)*(Hk*Hk)*H)); // Here it is Cteq^0.5
  } // End else
  double Cd = n*(Cdw+ Cdd + Cdl);
  // Ueq = 1.25/Hk*(Cf/2-((Hk-1)/(6.432*Hk)*((Hk-1)/(6.432*Hk)));
  bl->Cd[iPoint] = Cd;
  bl->Cf[iPoint] = Cf;
  bl->Delta[iPoint] = delta;
  bl->Hstar[iPoint] = Hstar;
  bl->Hstar2[iPoint] = Hstar2;
  bl->Hk[iPoint] = Hk;
  bl->Cteq[iPoint] = Cteq;
  bl->Us[iPoint] = Us;
}