#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## VII Coupler
# Paul Dechamps

from dart.viscous.Master.DAirfoil import Airfoil
from dart.viscous.Master.DWake import Wake
from dart.viscous.Drivers.DGroupConstructor import GroupConstructor

from fwk.coloring import ccolors

import numpy as np

class Coupler:
    def __init__(self, iSolver, vSolver) -> None:
        self.iSolver=iSolver
        self.vSolver=vSolver

        self.maxCouplIter = 2
        self.couplTol = 1e-4
        pass

    def CreateProblem(self, _boundaryAirfoil, _boundaryWake):

        self.group = [Airfoil(_boundaryAirfoil), Wake(_boundaryWake)] # airfoil and wake python objects
        pass

    def Run(self):

      # Initilize meshes in c++ objects
      
      couplIter = 0
      while couplIter < self.maxCouplIter:
        
        cdPrev = 0

        # Run inviscid solver

        self.iSolver.run()

        # Extract inviscid boundary

        for n in range(0, len(self.group)):

          for i in range(0, len(self.group[n].boundary.nodes)):

            self.group[n].v[i,0] = self.iSolver.U[self.group[n].boundary.nodes[i].row][0]
            self.group[n].v[i,1] = self.iSolver.U[self.group[n].boundary.nodes[i].row][1]
            self.group[n].v[i,2] = self.iSolver.U[self.group[n].boundary.nodes[i].row][2]
            self.group[n].Me[i] = self.iSolver.M[self.group[n].boundary.nodes[i].row]
            self.group[n].rhoe[i] = self.iSolver.rho[self.group[n].boundary.nodes[i].row]
        
        dataIn=[None,None]
        for n in range(0,len(self.group)):
          self.group[n].connectListNodes, self.group[n].connectListElems, dataIn[n]  = self.group[n].connectList()
        fMeshDict, cMeshDict, data = GroupConstructor().mergeVectors(dataIn, 0, 1)
        if couplIter == 0:
          # Initialize mesh          
          self.vSolver.InitMeshes(0, fMeshDict['locCoord'][:fMeshDict['bNodes'][1]], fMeshDict['globCoord'][:fMeshDict['bNodes'][1]])
          self.vSolver.InitMeshes(1, fMeshDict['locCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['globCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
          self.vSolver.InitMeshes(2, fMeshDict['locCoord'][fMeshDict['bNodes'][2]:], fMeshDict['globCoord'][fMeshDict['bNodes'][2]:])
          print('fro mpy : Mesh ok')
        self.vSolver.SetInvBC(0, fMeshDict['vtExt'][:fMeshDict['bNodes'][1]], fMeshDict['Me'][:fMeshDict['bNodes'][1]], fMeshDict['rho'][:fMeshDict['bNodes'][1]])
        self.vSolver.SetInvBC(1, fMeshDict['vtExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['Me'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['rho'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
        self.vSolver.SetInvBC(2, fMeshDict['vtExt'][fMeshDict['bNodes'][2]:], fMeshDict['Me'][fMeshDict['bNodes'][2]:], fMeshDict['rho'][fMeshDict['bNodes'][2]:])
        print('from py: InvBC ok')
        # Run viscous solver

        exitCode = self.vSolver.Run()
        print('From python exitcode: ', exitCode)

        error = (self.vSolver.cd - cdPrev)/self.vSolver.cd if self.vSolver.cd != 0 else 1

        if error  <= self.couplTol:
          print('')
          print(ccolors.ANSI_GREEN, 'It: {:<3.0f} Cd = {:<6.5f}. Error = {:<2.3f} < {:<2.3f}'
          .format(couplIter, self.vsolver.cd, np.log10(error), np.log10(self.couplTol)), ccolors.ANSI_RESET)
          print('')
          return 0
        
        print('')
        print(ccolors.ANSI_BLUE, 'It: {:<3.0f} Cd = {:<6.5f}. Error = {:<2.3f} > {:<2.3f}'
        .format(couplIter, self.vSolver.cd, np.log10(error), np.log10(self.couplTol)), ccolors.ANSI_RESET)
        print('')

        # Set inviscid boundary condition

        for n in range(0, len(self.group)):

          for i in range(0, self.group[n].nE):

            self.group[n].boundary.setU(i, self.group[n].u[i])

        couplIter += 1
      else:
        print(ccolors.ANSI_RED, 'Maximum number of {:<.0f} coupling iterations reached'.format(self.maxCouplIter), ccolors.ANSI_RESET)
        return couplIter

