################################################################################
#                                                                              #
#                            DARTFlo viscous module file                       #
#                    BIConditions: Initial and boundary conditions             #
#                        of the pseudo-time marching problem
#                                                                              #
# Author: Paul Dechamps                                                        #
# Université de Liège                                                          #
# Date: 2021.09.10                                                             #
#                                                                              #
################################################################################


# ------------------------------------------------------------------------------
#  Imports
# ------------------------------------------------------------------------------
from dart.viscous.Physics.DClosures import Closures

import matplotlib.pyplot as plt 
import numpy as np


class Initial:
    def __init__(self, BoundaryCond):
        self.boundCond=BoundaryCond
        self.mu_air=1.849e-5 # Dynamic viscosity of air at 25°C

    def initialConditions(self, var):
        """Twaithes integral solution of the (steady) Von Karman momentum
        integral equation for a laminar flow over a flat plate

        Parameters
        ----------
        - var: class
             Data container: - Solution
                             - Boundary layer parameters
                             - External flow parameters
                             - Mesh
                             - Transition information
        
        Tasks
        -----
        - Compute Twaithes solution of the momentum integral equation for a laminar flow over a flat plate in the
        regime prescibed by the inviscid solver. 
        - Define boundary conditions at the stagnation point and the first wake point 
        """

        
        u0=np.zeros(var.nN*var.nVar)        # Vector containing the intial solution
        
        # Loop over the three groups (upper, lower sides, wake)
        for n in range(3):

            if n==2:

                xx=var.xx[var.bNodes[n]:]                                         # Airfoil local coordinates.
                vt=var.extPar[var.bNodes[n]*var.nExtPar+2::var.nExtPar]           # Inviscid velocities.
                rho=var.extPar[var.bNodes[n]*var.nExtPar+1::var.nExtPar]          # Air density in each cell. 

            else:

                xx=var.xx[var.bNodes[n]:var.bNodes[n+1]]
                vt=var.extPar[var.bNodes[n]*var.nExtPar+2:var.bNodes[n+1]*var.nExtPar:var.nExtPar]
                rho=var.extPar[var.bNodes[n]*var.nExtPar+1:var.bNodes[n+1]*var.nExtPar:var.nExtPar]

            # Reynolds number based on the arc length.
            Rex=rho*vt*xx/self.mu_air
            Rex[Rex==0]=1 # Stagnation point
            Rex[Rex<=0] = 1
            ThetaTw=xx/np.sqrt(Rex)             # Momentum thickness.
            lambdaTw=np.zeros(len(ThetaTw))     # Dimensionless pressure gradient parameter.
            HTw=np.zeros(len(ThetaTw))          # Shape factor of the boundary layer.

            for iMarker in range(len(lambdaTw)):

                idxPrev = iMarker-1
                lambdaTw[iMarker] = (rho[iMarker] * ThetaTw[iMarker]**2) / self.mu_air * (vt[iMarker] - vt[idxPrev])#/(xx[idx]-xx[idxPrev])

                # Empirical correlations for H=f(lambda)
                if 0 <= lambdaTw[iMarker] < 0.1:

                    HTw[iMarker] = 2.61 - 3.75*lambdaTw[iMarker] + 5.24*lambdaTw[iMarker]**2

                elif -0.1<lambdaTw[iMarker]<0:

                    HTw[iMarker] = 2.088 + 0.0731 / (lambdaTw[iMarker] + 0.14)
            
            if n==2:

                u0[var.bNodes[n]*var.nVar+0::var.nVar]=ThetaTw      # Theta
                u0[var.bNodes[n]*var.nVar+1::var.nVar]=HTw          # H

            else:

                u0[var.bNodes[n]*var.nVar+0:var.bNodes[n+1]*var.nVar:var.nVar]=ThetaTw      # Theta
                u0[var.bNodes[n]*var.nVar+1:var.bNodes[n+1]*var.nVar:var.nVar]=HTw          # H

        u0[1::var.nVar]=2                           # H
        u0[2::var.nVar]=0                           # N
        u0[3::var.nVar]=var.extPar[2::var.nExtPar]  # Vt
        u0[4::var.nVar]=0.001                       # Ct
        
        # Set initial condition. 
        var.u=u0

        # Impose boudary conditions.
        self.boundCond.airfoilBC(var)
        self.boundCond.wakeBC(var)
        pass


class Boundaries:
    """ ---------- Boundary conditions ----------
        
        Boundary conditions at the stagnation point and on the first point of the wake.

        Attributes
        ----------
        Re: Flow Reynolds number.
        """

    def __init__(self, _Re):

        self.Re = _Re
        
    def airfoilBC(self, cfgFlow):
        """Boundary conditions at the stagnation point.
        Twaithes solution method values used for the
        stagnation point.
        """

        if cfgFlow.dv[0] > 0:
            T0 = np.sqrt(0.075 / (self.Re * cfgFlow.dv[0]))
        else:
            T0 = 1e-6

        H0  = 2.23                # One parameter family Falkner Skan.
        n0  = 0                   # Initial log of amplification factor.
        ue0 = cfgFlow.extPar[2]   # Inviscid flow velocity.
        Ct0 = 0                   # Stagnation point is laminar. 

        cfgFlow.blPar[cfgFlow.bNodes[0]*cfgFlow.nBlPar + 0] = ue0*T0*cfgFlow.Re       # Reynolds number based on the momentum thickness

        if cfgFlow.blPar[cfgFlow.bNodes[0]*cfgFlow.nBlPar+0] < 1:
            cfgFlow.blPar[cfgFlow.bNodes[0]*cfgFlow.nBlPar+0] = 1
            ue0 = cfgFlow.blPar[cfgFlow.bNodes[0]*cfgFlow.nBlPar + 0] / (T0*cfgFlow.Re)

        cfgFlow.blPar[cfgFlow.bNodes[0]*cfgFlow.nBlPar + 9] = T0 * H0                 # deltaStar

        if cfgFlow.xtrF == [0, 0]:
          # Turbulent closures.

          Ct0 = 0.03
          cfgFlow.blPar[cfgFlow.bNodes[0]*cfgFlow.nBlPar+1:cfgFlow.bNodes[0]*cfgFlow.nBlPar+9] = Closures.turbulentClosure(T0, H0,
                                                                                                    Ct0, cfgFlow.blPar[cfgFlow.bNodes[0]*cfgFlow.nBlPar+0],
                                                                                                    cfgFlow.extPar[cfgFlow.bNodes[0]*cfgFlow.nExtPar+0], 'wake')
        else:
          # Laminar closures.

          cfgFlow.blPar[(cfgFlow.bNodes[0]*cfgFlow.nBlPar + 1):(cfgFlow.bNodes[0]*cfgFlow.nBlPar+7)]=Closures.laminarClosure(T0, H0,
                                                                                                          cfgFlow.blPar[cfgFlow.bNodes[0]*cfgFlow.nBlPar+0],
                                                                                                          cfgFlow.extPar[cfgFlow.bNodes[0]*cfgFlow.nExtPar+0],
                                                                                                          'airfoil')
        
        # Set boundary condition.

        cfgFlow.u[cfgFlow.bNodes[0]*cfgFlow.nVar:(cfgFlow.bNodes[0]+1)*cfgFlow.nVar] = [T0, H0, n0, ue0, Ct0]
        pass

    # Wake boundary conditions 
    def wakeBC(self, cfgFlow):
        """Boundary conditions at the first point of the wake
        based on the parameters at the last point on the upper and lower sides.
        Drela (1989).
        """

        # Extract last points on the upper and lower sides

        xEnd_up = cfgFlow.u[(cfgFlow.bNodes[1]-1)*cfgFlow.nVar:cfgFlow.bNodes[1]*cfgFlow.nVar]
        xEnd_lw = cfgFlow.u[(cfgFlow.bNodes[2]-1)*cfgFlow.nVar:cfgFlow.bNodes[2]*cfgFlow.nVar]

        T0  = xEnd_up[0]+xEnd_lw[0]                                                # (Theta_up+Theta_low).
        H0  = (xEnd_up[0]*xEnd_up[1]+xEnd_lw[0]*xEnd_lw[1])/T0                     # ((Theta*H)_up+(Theta*H)_low)/(Theta_up+Theta_low).
        n0  = 9
        ue0 = cfgFlow.extPar[cfgFlow.bNodes[2]*cfgFlow.nExtPar+2]                  # Inviscid velocity.
        Ct0 = (xEnd_up[0]*xEnd_up[4]+xEnd_lw[0]*xEnd_lw[4])/T0                     # ((Ct*Theta)_up+(Theta)_low)/(Ct*Theta_up+Theta_low).

        cfgFlow.blPar[cfgFlow.bNodes[2]*cfgFlow.nBlPar+0] = ue0*T0*cfgFlow.Re      # Reynolds number based on the momentum thickness.
        
        if cfgFlow.blPar[cfgFlow.bNodes[2]*cfgFlow.nBlPar+0] < 1:
            cfgFlow.blPar[cfgFlow.bNodes[2]*cfgFlow.nBlPar+0] = 1
            ue0 = cfgFlow.blPar[cfgFlow.bNodes[2]*cfgFlow.nBlPar+0] / (T0*cfgFlow.Re)

        cfgFlow.blPar[cfgFlow.bNodes[2]*cfgFlow.nBlPar+9] = T0*H0                  # deltaStar
        
        # Turbulent closures.

        cfgFlow.blPar[cfgFlow.bNodes[2]*cfgFlow.nBlPar+1:cfgFlow.bNodes[2]*cfgFlow.nBlPar+9] = Closures.turbulentClosure(T0, H0,
                                                                                                    Ct0, cfgFlow.blPar[cfgFlow.bNodes[2]*cfgFlow.nBlPar+0],
                                                                                                    cfgFlow.extPar[cfgFlow.bNodes[2]*cfgFlow.nExtPar+0], 'wake')

        # Set boundary condition. 

        cfgFlow.u[cfgFlow.bNodes[2]*cfgFlow.nVar:(cfgFlow.bNodes[2]+1)*cfgFlow.nVar] = [T0, H0, n0, ue0, Ct0]
        pass